import React, {
  useRef,
  useState,
  useEffect,
  useImperativeHandle,
} from 'react';
import * as Three from 'three';
import { useFrame, useThree } from '@react-three/fiber';
import { HoloVideoObjectThreeJS } from '@MixedRealityCaptureStudios/holo-video-object-three';

const tuple = <T extends any[]>(...args: T): T => args;

type Vec3 = [number, number, number];
const multiply = ([ax, ay, az]: Vec3, [bx, by, bz]: Vec3): Vec3 => (
  [ax * bx, ay * by, az * bz]
);

// The IntrinsicElements['mesh'] type apparently takes a number, Vec3,
// or Three.Vector3 for the scale property, we'll narrow this type
// into a Vec3 to do our resizing logic.
const getScale = (scale: number | Vec3 | Three.Vector3): Vec3 => {
  if (typeof scale === 'number') {
    return tuple(scale, scale, scale);
  }

  if (scale instanceof Three.Vector3) {
    return tuple(scale.x, scale.y, scale.z);
  }

  return scale ?? tuple(1, 1, 1);
};

type HoloVideoObjectProps = {
  path: string;
  playing?: boolean;
  muted?: boolean;
  onUpdate?: (currentTime: number, totalTime: number) => void;
  useFileCoordinates?: boolean;
  meshProps?: JSX.IntrinsicElements['mesh'];
};

type Props = HoloVideoObjectProps;
type Ref = HoloVideoObjectThreeJS;
export const HoloVideoObject = React.forwardRef<Ref, Props>((props, ref) => {
  const { gl } = useThree();
  const [hvoMesh, setHvoMesh] = useState<Three.Mesh>();
  const hvoRef = useRef<HoloVideoObjectThreeJS>();
  useImperativeHandle(ref, () => hvoRef.current);

  const {
    path,
    playing = false,
    muted = false,
    useFileCoordinates = false,
    onUpdate = () => {},
    meshProps,
  } = props;

  // Synchronize our play state.
  useEffect(() => {
    if (!hvoRef.current) {
      return;
    }

    if (playing) {
      hvoRef.current.play();
    } else {
      hvoRef.current.pause();
    }
  }, [playing, hvoRef.current?.hvo.id]);

  useEffect(() => {
    if (!hvoRef.current) {
      return;
    }

    hvoRef.current.setAudioEnabled(!muted);
  }, [muted]);

  useFrame(() => {
    if (!hvoRef.current) {
      return;
    }

    hvoRef.current.update();
    const currentTime = hvoRef.current.hvo.lastVideoTime;
    const videoDuration = hvoRef.current.hvo?.videoElement?.duration ?? 0;

    // For some reason the duration can be NaN on the video element here.
    // TODO(Jordan): We should fix this when we add the duration to our public API.
    const adjustedDuration = Number.isNaN(videoDuration) ? 0 : videoDuration
    const totalTime = adjustedDuration * 1000;
    onUpdate(currentTime, totalTime);
  });

  useEffect(() => {
    const hvo = new HoloVideoObjectThreeJS(gl, (mesh) => { setHvoMesh(mesh); });

    hvo.open(path, { autoloop: true, audioEnabled: true, keepAllMeshesInMemory: true });

    hvoRef.current = hvo;

    return () => { hvoRef.current.close(); };
  }, [props.path]);

  const scale = getScale(meshProps.scale);

  const centimetersToMeters = tuple(0.01, 0.01, 0.01);

  // Three doesn't have an exact unit of distance,
  // however the geometry primitives use a default size of 1.
  // So by default we'll scale down our file units.
  const adjustedScale = useFileCoordinates
    ? scale
    : multiply(scale, centimetersToMeters);

  const adjustedMeshProps = {
    position: tuple(0, 0, 0),
    ...meshProps,
    scale: adjustedScale,
  };

  return hvoMesh && (
    <mesh
      {...adjustedMeshProps}
      castShadow
      receiveShadow
      geometry={hvoMesh.geometry}
      material={hvoMesh.material}
    />
  );
});