# Holo Video Object

holo-video-object is a library for holographic video playback on the web.

## Installation

`npm i holo-video-object`

## Quick Start

```ts
import { HoloVideoObject } from 'holo-video-object';

// Create a WebGL context & buffers.
const gl = canvas.getContext('webgl2');

const position = gl.createBuffer();
const index = gl.createBuffer();
const uv = gl.createBuffer();
const normal = gl.createBuffer();
const texture = gl.createTexture();

const hvo = new HoloVideoObject(gl);

hvo.onLoaded = (fileInfo) => {
    // hvo will fill these buffers on each call to `updateBuffers`.
    hvo.setBuffers(position, index, uv, normal, texture);

    hvo.updateBuffers();

    // Invoke custom drawing behavior using the buffers we provided to hvo.
    // See the "Starting playback" section for more information about
    // animating our hologram.
    drawScene();
};

hvo.open(url, { autoloop: true });
```

## Adding HCap content to the scene

Options available in open():

Name | Description | Default
---|--|--
audioEnabled | Specifies whether audio playback is enabled initially | false
autoloop | sets file to autoloop | false
autoplay | Specifies whether capture should automatically begin playback when loaded. If enabled will disable audio as many browsers do not allow automatic audio playback without user input. | false
minBuffers|Minimum number of .bin segments that must be buffered before capture is considered preloaded.|2
maxBuffers|Maximum number of .bin segments that can be buffered and kept in memory at one time.|3
streamMode|Force the "streamMode" used for video texture streaming. This should only be used for debugging.<ul><li>HoloVideoObject.StreamMode.Automatic</li><li>HoloVideoObject.StreamMode.MP4 (downloads entire mp4 before beginning playback)</li><li>HoloVideoObject.StreamMode.HLS (generally used by Safari)</li><li>HoloVideoObject.StreamMode.Dash (Downloads portions of video in chunks, allowing earlier playback)</li><lu>Not all modes work for all browsers. Generally, the Automatic default is best. If you discover a browser for which Automatic does not result in the best performance, please alert Microsoft MRCS.|Automatic
keepAllMeshesInMemory | Forces all .bin segments to be kept in memory. This overrides the maxBuffers setting to the total number of segments. | false
startTime | Start playback from specified time (in seconds) | 0

## Starting playback

When the open callback is invoked, the `HoloVideoObject` instance will have at least one preview frame loaded but it may require additional time to buffer the rest of the capture. The application should monitor the `state` property of the `HoloVideoObject` instance until it reaches the value of `State.Opened`, at which point playback can be started.

Many browsers will only allow video playback to be initiated in response to user input, so to begin playback the application should call the `HoloVideoObject.play()` method from a user input event handler as shown below:

~~~~ts
renderer.domElement.addEventListener('mousedown', function() {
if (hvo.state === States.Opened ||
    hvo.state === States.Opening) {
    hvo.play();
}});
~~~~

## Displaying animated HCap content

In order to update the hologram instance to display the texture and geometry for the current frame of playback, the `updateBuffers` method needs to be called from the application's animation tick method as shown below:

~~~~ts
function animate() {
    requestAnimationFrame(animate);

    // Update hologram to latest frame of playback:
    hvo.updateBuffers();
}
~~~~

Building a pure WebGL app with HCap content
============

The application is responsible for allocating the WebGL texture that receives the capture texture atlas, as well as index and vertex buffers for each vertex component (position, uv coordinate, and optionally vertex normals) used to display the capture and providing them to the `HoloVideoObject` instance via the `setBuffers` API.

A full WebGL sample application is provided in the `mrcs-examples` package.

## DASH streaming

MPEG-DASH is the preferred streaming format for HCap content on browsers other than Safari. DASH playback requires the `dash.js` player script from https://github.com/Dash-Industry-Forum/dash.js. The following line can be used to include the `dash.js` player implementation in a web page:
~~~~html
<script src="https://cdn.dashjs.org/v4.4.1/dash.mediaplayer.min.js"></script>
~~~~

## Known Limitations

- An apparent bug has been identified where the HLS video stream used by default on iOS and macOS will not play correctly in Safari 14 on Apple Silicon/M1 systems. Because it isn't
possible for the plugin to distinguish an M1 system from an Intel system, the plugin now resolves `StreamMode.Automatic` to `StreamMode.MP4` (instead of `StreamMode.HLS`) in Safari 14 on *all* macOS systems to ensure maximum out of the box compatibility. As such the application developer may prefer instead to give macOS users the ability to choose a specific format based on their system, or manually specify `StreamMode.HLS` on macOS and advise M1 users to use Safari 15 instead.

- Capture audio will always be muted by default on Safari, as Safari will not pre-load video data unless it is muted. The client application should call `HoloVideoObject.setAudioEnabled(true)` before starting playback if audio playback is desired.
