import { HoloVideoObject, States, ErrorStates } from "./holovideoobject";
import { FileInfo, OpenOptions, HoloVideoCreateOptions } from "./holoVideoInterface";

export abstract class HoloVideoObjectProxy {
    onEndOfStream: any;
    hvo: HoloVideoObject;
    state: States;
    fileInfo: FileInfo;

    private _hvoOnEndOfStream(hvo) {
        if (this.onEndOfStream) {
            this.onEndOfStream(this);
        }
    }

    protected abstract _createHoloVideoObject(
        createOptions?: HoloVideoCreateOptions,
        updateCurrentFrameCallback?: (frame: number) => void,
        errorCallback?: (type: ErrorStates, info?: Error | MediaError) => void
    ): HoloVideoObject;

    constructor(
        callback: () => void,
        createOptions?: HoloVideoCreateOptions,
        updateCurrentFrameCallback?: (frame: number) => void,
        errorCallback?: (type: ErrorStates, info?: Error | MediaError) => void
    ) {
        const hvo = this._createHoloVideoObject(createOptions, updateCurrentFrameCallback, errorCallback);

        this.hvo = hvo;
        hvo.onEndOfStream = this._hvoOnEndOfStream.bind(this);
        hvo.onUpdateCurrentFrame = updateCurrentFrameCallback;
        hvo.onLoaded = (fileInfo) => {
            this.fileInfo = fileInfo;
            this.state = this.hvo.state;
            callback();
        };
    }

    /**
     * Opens and capture and begins buffering data for playback
     *
     * @param {string} url - Capture URL, e.g. "http://127.0.0.1/mycapture.hcap"
     * @param {Object} options - Optional collection of options related to capture loading and playback.
     * @param {boolean} options.audioEnabled - Specifies whether audio playback is enabled initially
     * @param {boolean} options.autoplay - Specifies whether capture should automatically begin playback when loaded. If enabled will disable audio as many browsers do not allow automatic audio playback without user input.
     * @param {HoloVideoObject.StreamMode} options.streamMode - Force the streamMode for the video object. Default is Automatic.
     * @param {number} options.minBuffers - Minimum number of .bin segments that must be buffered before capture is considered preloaded. Default value is 2.
     * @param {number} options.maxBuffers - Maximum number of .bin segments that can be buffered and kept in memory at one time.
     * @param {boolean} options.keepAllMeshesInMemory - Overrides maxBuffers setting and keeps all mesh in memory to prevent further loading.
     * @param {number} options.startTime - Time in seconds where playback should begin from. Defaults to 0 if not specified.
     * @param {number} options.crossOrigin - Value that will be passed through to the `crossOrigin` propety on the plugin's internal `VideoElement`. The default value is "use-credentials".
     */
    public open(url: string, options: OpenOptions): void {
        if (this.state > States.Empty) {
            this.close();
        }
        this.hvo.open(url, options);
        this.state = this.hvo.state;
    }

    /**
     * Updates capture mesh and texture to latest playback frame. This should be called periodically from the application's animate/update loop.
     */
    public update(): void {
        if (this.hvo) {
            this.state = this.hvo.state;
        }

        this.hvo.update();
    }

    /**
     * Rewinds capture back to the first frame and pauses.
     */
    public rewind(): void {
        this.hvo.rewind();
    }

    /**
     * Seeks playback position to the specified time.
     * @param {number} seekTime - Time (in seconds) to seek to.
     * @param {boolean} displayImmediately - Specifies whether to update to and display target frame immediately.
     */
    public seekToTime(seekTime: number, displayImmediately: boolean): void {
        this.hvo.seekToTime(seekTime, displayImmediately);
    }

    /**
     * Starts capture playback if sufficient data has been buffered, or resumes paused playback.
     */
    public play(): void {
        if (this.hvo.state == States.Opening) {
            this.hvo.forceLoad();
        } else if (this.hvo.state >= States.Opened && this.hvo.state != States.Playing) {
            this.hvo.play();
        }
    }

    /**
     * Stops playback and releases capture-specific resources. A new capture can then be loaded by calling {@link HoloVideoObjectProxy#open}
     */
    public close(): void {
        this.hvo.close();
    }

    /**
     * Pauses playback.
     */
    public pause(): void {
        this.hvo.pause();
    }

    /**
     * Sets verbosity level of log output
     * @param {number} level - 0 = errors, 1 = warnings, 2 = info, 3 = debug. Default setting is = 0.
     */
    public setLogLevel(level: number): void {
        this.hvo.logLevel = level;
    }

    /**
     * Enables or disables audio playback. May be called at any time.
     * @param {boolean} enabled - Specifies whether audio should be enabled or disabled.
     */
    public setAudioEnabled(enabled: boolean): void {
        this.hvo.setAudioEnabled(enabled);
    }

    /**
     * Returns whether audio playback is currently enabled.
     */
    public audioEnabled(): boolean {
        return this.hvo.audioEnabled();
    }

    /**
     * Sets audio volume.
     * @param {number} volume - Value between 0 and 1 specifying volume level.
     */
    public setAudioVolume(volume: number): void {
        this.hvo.setAudioVolume(volume);
    }

    /**
     * Specifies whether capture should loop automatically. May be called at any time.
     * @param {boolean} loop
     */
    public setAutoLooping(loop: boolean): void {
        this.hvo.setAutoLooping(loop);
    }
}
